; *************************************************************************************************
; *
; *	Title:	TIMER.ASM
; *	Copyright (c) March 1994, Ryu Consulting
; *	Written by Rahner James
; *
; *	This file contains the driver for the timer interrupt
; *
; *************************************************************************************************

TIMER_INT		equ	8
FLAG_IN_ISR		equ	0		; Set if in the timer ISR



	include	driver.inc

Driver_Name		db	"Timer", 0		; Name of the driver


; *************************************************************************************************
; *
; *				Data relative to this driver
; *
; *************************************************************************************************

	even
Jump_Table label word
	dw	init_driver
	dw	stop_driver

	db	50 dup('Hay')
	db	"Needle"
	db	50 dup('Hay')
Local_Stack label word

	db	50 dup('Hay')
	db	"Needle"
	db	50 dup('Hay')
ISR_Stack label word
ISR_SP	dw	0
ISR_SS	dw	0

Old_SP	dw	0
Old_SS	dw	0

Old_Timer_Vector	dd	0	; -> old timer vector
Timer_Bits		dw	0	; B0: timer semaphore, 1=in timer ISR


; *************************************************************************************************
; *
; *	STRING_OUT
; *	Outputs an ASCIIz string to the console
; *
; *	Given:
; *		CS:SI -> string to output
; *
; *	Returns:
; *		SI -> 0 terminator
; *
; *************************************************************************************************
string_out proc

	pop	si
@@:	mov	dl, cs:[si]
	inc	si
	or	dl, dl
	jz	short done

	mov	ah, 6
	int	21h
	jmp	@B

done:	jmp	si

string_out endp


; *************************************************************************************************
; *
; *	INIT_DRIVER
; *	Initialize this driver
; *
; *	Given:
; *		AX = 0
; *		SS = CS
; *		DS = ES = global data segment
; *
; *	Returns:
; *		Driver initialized, no registers (except CS,SS,SP) preserved
; *
; *************************************************************************************************
init_driver proc

ifdef DEBUG
	STROUT	"Initializing timer driver ..."
	CRLF
endif
; *
; * See if the INT is already revectored
; *
	xor	eax, eax

	mov	ds:ET_S.timer_semaphore, 1
	mov	ds:ET_S.tick_count, eax
	mov	ds:ET_S.first_timer, eax
	mov	ds:ET_S.total_timer_events, ax
	mov	ds:ET_S.timer_semaphore, al

	cmp	cs:Old_Timer_Vector, eax	; See if we are already installed
	jnz	short done

	mov	es, ax
	mov	ax, cs
	shl	eax, 16
	mov	ax, offset timer_isr
	cli
	xchg	es:[TIMER_INT*4], eax
	mov	cs:Old_Timer_Vector, eax
	sti

done:	ret

init_driver endp


; *************************************************************************************************
; *
; *	STOP_DRIVER
; *	Stops this driver
; *
; *	Given:
; *		AX = 1
; *		SS = CS
; *		DS = ES = global data segment
; *
; *	Returns:
; *		Driver stopped, all INTs returned, no registers (except CS,SS,SP) preserved
; *
; *************************************************************************************************
stop_driver proc

ifdef DEBUG
	STROUT	"Stopping timer driver ..."
	CRLF
endif

; *
; * See if the INT was revectored
; *
	xor	ecx, ecx
	mov	es, cx
	cli
	xchg	cs:Old_Timer_Vector, ecx	; Xchg with the old timer vector
	jecxz	short done			; Skip if we were not
	mov	es:[TIMER_INT*4], ecx

done:	sti
	ret

stop_driver endp


; *************************************************************************************************
; *
; *	TIMER_ISR
; *	Interrupt Service Routine for the timer tick
; *
; *	Given:
; *		CS = this segment
; *
; *	Returns:
; *		Timer variables and structures updated
; *
; *************************************************************************************************
timer_isr proc far

	push	ds
	mov	ds, word ptr cs:Global_Data_Ptr+2
	inc	ds:ET_S.tick_count			; Bump the counter regardless of what else is happening

	bts	cs:Timer_Bits, FLAG_IN_ISR		; Set/check our semaphore
	jc	short done
	sti

; *
; * Use our ISR stack
; *
	push	eax
	mov	cs:ISR_SS, ss				; Save SS:SP
	mov	cs:ISR_SP, sp
	mov	ax, cs
	mov	ss, ax
	mov	sp, offset ISR_Stack

	cmp	ds:ET_S.timer_semaphore, 0		; See if the foreground process is messing with it
	jnz	short popds_done

; *
; * Save and setup our registers
; *
	push	es
	mov	es, word ptr cs:Global_Data_Ptr+2
	push	ebx
	push	ecx
	push	edx
	push	edi
	push	esi

; *
; * Restore the registers used
; *
	pop	esi
	pop	edi
	pop	edx
	pop	ecx
	pop	ebx
	pop	es

popds_done:
	lss	sp, dword ptr cs:ISR_SP
	pop	eax
	and	cs:Timer_Bits, not (1 shl FLAG_IN_ISR)		; Clear our semaphore

done:	pop	ds
	jmp	dword ptr cs:[Old_Timer_Vector]

timer_isr endp


; *************************************************************************************************
; *
; *	START_OF_CODE
; *	This is the entry point from the protected-mode client
; *
; *	Given:
; *		AX = task to perform,
; *			0 = initialization
; *			1 = disengage
; *
; *	Returns:
; *		Jumps to next in line after completion
; *		All registers can be destroyed except EAX, SS, SP
; *
; *************************************************************************************************
start_of_code:
	push	ax
	mov	cs:Old_SS, ss
	mov	cs:Old_SP, sp

	cmp	ax, 2			; See if too big
	jnc	short @F

	cld
	sti
	mov	ds, word ptr cs:Global_Data_Ptr+2
	mov	es, word ptr cs:Global_Data_Ptr+2
	mov	bx, cs
	mov	ss, bx
	mov	sp, offset Local_Stack

	mov	bx, ax
	add	bx, bx			; BX *= 2
	call	cs:Jump_Table[bx]

@@:	lss	sp, dword ptr cs:Old_SP
	pop	ax
	jmp	dword ptr cs:[Next_Driver]


	end


